import json
import logging
import pickle
import re
import warnings
from base64 import b64decode, b64encode
from collections import defaultdict
from copy import deepcopy
from functools import lru_cache

import numpy as np
import pandas as pd
from numpy.lib import recfunctions as rfn
from numtools.serializer import Serializer
from numtools.misc import replace_nan

from nastranio.constants import ELEMENT, PROPERTY, UNKNOWN
from nastranio.decorators import cached_property
from nastranio.fields_writer import (
    DefaultDict,
    fields_to_card,
    get_field,
    nbrows_by_fields,
    trans,
)
from nastranio.pylib import autoconvert
from nastranio.utils import array2dic, calcdiff, dic2array, transform_dict_of_list

try:
    import msgpack

    ISMSGPACK = True
except ImportError:
    ISMSGPACK = False

pat = re.compile(r"(?P<root>\w+)(?P<id>\d+)")

# ============================================================================
# json encoder / decoder
# ============================================================================


class PythonObjectEncoder(json.JSONEncoder):
    """encode sets to json"""

    def default(self, obj):
        if isinstance(obj, (list, dict, str, int, float, bool, type(None))):
            return super().default(obj)
        return {"_python_object": b64encode(pickle.dumps(obj)).decode("utf-8")}


def as_python_object(dct):
    if "_python_object" in dct:
        return pickle.loads(b64decode(dct["_python_object"].encode("utf-8")))
    return dct


# ============================================================================
# msgpack encoder / decoder
# ============================================================================


def msgpack_encode(obj):
    if isinstance(obj, set):
        obj = {"__set__": True, "as_list": list(obj)}
    return obj


def msgpack_decode(obj):
    if "__set__" in obj:
        obj = set(obj["as_list"])
    return obj


# ============================================================================
# parsing helpers
# ============================================================================


def parse_table(table, asruler=False, linesno=None):
    r"""parse a Nastran user-guide style table,
    and return a tuple (`card`, `fields`, `repeated`)

    >>> table = (
    ...    '| 1    | 2   | 3   | 4   | 5  | 6  | 7  | 8   | 9  | 10 |\n'
    ...    '|------+-----+-----+-----+----+----+----+-----+----+----|\n'
    ...    '| PBAR | PID | MID | A   | I1 | I2 | J  | NSM |    |\n'
    ...    '| ""   | K1  | K2  | I12 |    |    |    |     |    |\n')
    >>> card, fields, repeated = parse_table(table)
    >>> card
    'PBAR'
    >>> fields
    {2: 'PID', 3: 'MID', 4: 'A', 5: 'I1', 6: 'I2', 7: 'J', 8: 'NSM', 12: 'K1', 13: 'K2', 14: 'I12'}
    >>> repeated  # -> None

    It also handles tables using -etc.- fields:

    >>> table = (
    ...    '| 1     | 2    | 3  | 4      | 5     | 6      | 7    | 8      | 9     | 10 |\n'
    ...    '|-------+------+----+--------+-------+--------+------+--------+-------+----|\n'
    ...    '| PCOMP | PID  | Z0 | NSM    | SB    | FT     | TREF | GE     | LAM   | \n'
    ...    '| ""    | MID1 | T1 | THETA1 | SOUT1 | MID2   | T2   | THETA2 | SOUT2 | \n'
    ...    '| ""    | MID3 | T3 | THETA3 | SOUT3 | -etc.- |      |        |       | \n')
    >>> card, fields, repeated = parse_table(table)
    >>> card
    'PCOMP'
    >>> fields
    {2: 'PID', 3: 'Z0', 4: 'NSM', 5: 'SB', 6: 'FT', 7: 'TREF', 8: 'GE', 9: 'LAM'}
    >>> repeated
    {'starts@': 12, 'fields': ['MID', 'T', 'THETA', 'SOUT']}

     -etc.- is not supposed to be trailing:

    >>> table = (
    ...    '| 1    | 2    | 3   | 4   | 5      | 6      | 7   | 8   | 9   | 10 |\n'
    ...    '|------+------+-----+-----+--------+--------+-----+-----+-----+----|\n'
    ...    '| RBE2 | EID  | GN  | CM  | GM1    | GM2    | GM3 | GM4 | GM5 |    | \n'
    ...    '| ""   | GM6  | GM7 | GM8 | -etc.- | ALPHA  |     |     |     |    | \n')
    >>> card, fields, repeated = parse_table(table)
    >>> card
    'RBE2'
    >>> fields
    {2: 'EID', 3: 'GN', 4: 'CM'}
    >>> repeated
    {'starts@': 5, 'fields': ['GM'], 'floating': ['ALPHA']}
    """
    lines = [l.strip() for l in table.split("\n")]
    lines = [l for l in lines if l]
    # ---------------------------------------
    # ensure first line is field numbering
    if set(lines[0]) == {" ", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "|"}:
        lines = lines[1:]
    # ---------------------------------------
    # check if second line is grid decoration
    if set(lines[0]) == {"+", "-", "|"}:
        lines = lines[1:]
    if linesno:
        lines = [l for i, l in enumerate(lines) if i in linesno]
    # ---------------------------------------
    # get fields
    fields = {}
    for i, line in enumerate(lines):
        _fields = line.split("|")
        # since grid has trailing and heading "|" characters, we need to remove first and last fields
        _fields = [f.strip() for f in _fields[1:-1]]
        fields.update(dict(zip(range(1 + i * 10, len(_fields) + 1 + i * 10), _fields)))
    # ---------------------------------------
    # clean fields: remove empty fields
    for k, v in fields.copy().items():
        if v.strip() in ("", '""'):
            fields.pop(k)
    if asruler:
        # --------------------------------------------------------------------
        # ensure all fields length < 8 characters
        for k, fieldname in fields.items():
            if len(fieldname) > 6:
                fields[k] = fieldname[:5] + ".."
        repeated = {}
    elif "-etc.-" in fields.values():
        etc_key = {v: k for k, v in fields.items()}["-etc.-"]
        floating = list({k: v for k, v in fields.items() if k > etc_key}.values())
        fields = {k: v for k, v in fields.items() if k < etc_key}
        # repeated fields can be one or more.
        # iterate from last field until we find a repeted pattern
        i = -1
        keys = list(fields.keys())
        repeated = {"starts@": None, "fields": []}
        while True:
            key = keys[i]
            value = fields[key]
            field, nb = pat.search(value).groups()
            if field in repeated["fields"]:
                break
            repeated["fields"].append(field)
            i -= 1
        repeated["fields"] = repeated["fields"][::-1]
        # clean all the repeated fields
        for key in keys[::-1]:
            value = fields[key]
            try:
                field, nb = pat.search(value).groups()
            except:
                break
            else:
                fields.pop(key)
                repeated["starts@"] = key
        if floating:
            repeated["floating"] = floating
    else:
        repeated = None
    if 1 in fields:
        card = fields.pop(1)
    else:
        card = None
    return card, fields, repeated


class DummyFields:
    """drop-in replacement for fields dict. Used by anonymous cards"""

    def __getitem__(self, x):
        return "FIELD#%d" % x


# ============================================================================
# Cards Skeletons to be derived
# ============================================================================


class SimpleCard(Serializer):
    """defines generically simple cards (eg. GRID, CQUAD4),
    or cycling cards (eg. PCOMP)
    """

    TABLE = None
    DEFAULTS = {}
    MULT_TYPES_FIELDS = {}
    type = UNKNOWN

    def __init__(self, name=None, data=None):
        if not name:
            name = self.__class__.__name__
        self.name = name
        if hasattr(self, "TABLE") and self.TABLE is not None:
            self.card, self.fields, self.repeated = parse_table(self.TABLE)
        else:
            self.card = name
            self.fields = {}  # DummyFields()
            self.repeated = None
        # --------------------------------------------------------------------
        # main property (existing for all cards)
        self.carddata = {"main": defaultdict(list)}
        if data:
            self.resume(data)
        # if _is_dummy:
        #     # build fields based on field number
        #     _fields = dict(self.carddata["main"])
        #     fields = dict(zip(range(0, len(_fields)), _fields.keys()))
        #     self.fields = {
        #         k: v for k, v in fields.items() if (k % 10 != 0 and (k - 1) % 10 != 0)
        #     }

    @classmethod
    def fields_info(cls):
        _data = parse_table(cls.TABLE)
        all_fields = tuple(_data[1].values())
        repeated = _data[2]
        fields = {
            # mandatory needs to keep order
            "mandatory": tuple([f for f in all_fields if f not in cls.DEFAULTS]),
            "optional": set([f for f in all_fields if f in cls.DEFAULTS]),
        }
        if repeated:
            fields["repeated"] = tuple([f"{field}i" for field in repeated["fields"]])
        return fields

    @cached_property
    def _rowno2id(self):
        """map carddata row number to card IDs"""
        return {i: id for i, id in enumerate(self.carddata["main"][self.XID_FIELDNAME])}

    @cached_property
    def _id2rowno(self):
        """map card IDs to carddata row number"""
        return {id: i for i, id in enumerate(self.carddata["main"][self.XID_FIELDNAME])}

    def _list_caches(self):
        return {k for k in self.__dict__ if k.startswith(cached_property.CACHED_PREFIX)}

    def clear_caches(self, rebuild=True):
        """delete internal caches"""
        prefix = cached_property.CACHED_PREFIX
        cache_names = self._list_caches()  # _cache_XXX
        cached_names = [k.replace(prefix, "") for k in cache_names]  # XXX
        logging.info("clean cached properties: %s", ", ".join(cached_names))
        _rebuilt = []
        for fcached, cache_name in zip(cached_names, cache_names):
            self.__dict__.pop(cache_name)
            if rebuild:
                # fcached = cached_name.replace(prefix, "")
                getattr(self, fcached)
                _rebuilt.append(fcached)
        if _rebuilt:
            logging.info("rebuilt cache: %s", ", ".join(_rebuilt))

    def default_to_nan(self, transfields=True):
        def _default_to_nan(row):
            """change default values to None"""
            for k, v in row.items():
                v = self._check_default(k, v, None)
                if transfields:
                    v = trans(v)
                row[k] = v
            return row

        df = pd.DataFrame(self.array)
        df = df.astype(object).apply(_default_to_nan, axis=1)
        if transfields:
            df = df.fillna("")
        return df

    def diff(self, other):
        return calcdiff((self.__dict__, other.__dict__))

    def __eq__(self, other):
        return len(self.diff(other)) == 0

    def _reset(self):
        name = self.name
        self.__dict__ = {}
        self.__init__(name=name)

    # def __getattr__(self, attr):
    #     """bind carddata contents"""
    #     if attr in self.carddata:
    #         return self.carddata[attr]
    #     raise AttributeError(attr)

    def __len__(self):
        keys = list(self.carddata["main"].keys())
        if len(keys) == 0:
            return 0
        key = keys[0]
        return len(self.carddata["main"][key])

    def help(self, doprint=False):
        """
        return self.TABLE usable as ruler
        """
        card, fields, repeated = parse_table(self.TABLE, asruler=True)
        # substitute fields keys with appropriate placeholder
        ruler_fieldsmap = DefaultDict(fn1=card)
        ruler_fieldsmap.update({"fn%d" % k: field for k, field in fields.items()})
        lines = fields_to_card(ruler_fieldsmap, leading="$", sep="▕")
        if not doprint:
            return lines
        print("\n".join(lines))

    def _check_default(self, key, value, default=None):
        if default is None:
            default = self.DEFAULTS
        if key in default and value == default[key]:
            value = None
        return value

    def _to_nastran(self, ids=frozenset(), ruler=False, additional_ruler_lines=()):
        """
        common code for all `to_nastran` functions
        """
        # create a DefaultDict dict mapping placeholders "fnXX" to field names
        # {'fn1': 'CORD2R', 'fn2': 'CID', ..., 'fn12': 'C1', 'fn13': 'C2', 'fn14': 'C3', }
        fieldsmap = DefaultDict(fn1=self.card)  # , fn10='+', fn11='+')
        try:
            fieldsmap.update({"fn%d" % k: v for k, v in self.fields.items()})
        except:
            print(self.fields)
            raise
        lines = []
        # creating a ruler is as simple as:
        if ruler:
            lines += self.help()
            if additional_ruler_lines:
                lines += additional_ruler_lines
        # ---------------------------------------------------------------
        # associate carddata['main'] entries to fields
        field = self.fields[2]
        all_ids = self.carddata["main"][field]  # eg. [eid1, eid2, ...]
        ids = set(ids) & set(all_ids)
        if not ids:
            ids = set(all_ids)
        # associate card number in the list to card ID
        full_array = self.array
        _isin = np.isin(full_array[field], list(ids))
        array = full_array[_isin]
        ixs = np.flatnonzero(_isin)
        return fieldsmap, lines, zip(ixs, array[field])

    def _newline(self, fieldsmap, ix, with_default):
        # build a new row
        data = DefaultDict(fn1=self.card)
        for fieldcode, fieldname in fieldsmap.items():
            if fieldcode in ("fn1",) or fieldname in ("+"):
                continue
            value = replace_nan(self.carddata["main"][fieldname][ix])
            if not with_default:
                value = self._check_default(fieldname, value, default=self.DEFAULTS)
            data[fieldcode] = trans(value)
        return data

    def to_nastran(
        self, ids=frozenset(), ruler=False, with_default=True, comments=None
    ):
        """
        Convert a collection of cards to 8-chars NASTRAN format.
        SimpleCard to_nastran()"""
        fieldsmap, lines, ixid = self._to_nastran(ids=ids, ruler=ruler)
        for ix, id in ixid:
            # initiate a new line and populate a data dict with fixed fields
            try:
                data = self._newline(fieldsmap, ix, with_default)
            except ValueError as exc:
                logging.critical(f"cannot make new line for {self.name} {ix}")
                raise
            if comments and hasattr(self, "COMMENTS_KEY"):
                comment = comments.get(self.COMMENTS_KEY, dict()).get(id, dict())
                if comment:
                    lines.append(f"$ Femap {self.COMMENTS_KEY} {id} : {comment}")
            lines += fields_to_card(data)
        return lines

    def merge(self, other):
        """attempt to create a universal merge() method working for
        SimpleCard, SimpleCyclingCard and ComplexCard
        """
        # --------------------------------------------------------------------
        # merge all fields but the extended data IDs
        assert self.name == other.name
        # count for later checks
        nb_ini = len(self)
        nb_to_add = len(other)
        for k, v in other.carddata["main"].items():
            # avoid reference fields
            dataname, *_ = k.split("ID")
            if dataname in self.carddata:
                continue
            self.carddata["main"][k] += v
        # --------------------------------------------------------------------
        # merge extended data
        for dataname in self.carddata.keys():
            if dataname == "main":
                continue
            colname = "%sID" % dataname
            offset = max(self.carddata["main"][colname]) + 1
            other._offset_repeated_ix(colname, offset=offset)
            sd = self.carddata[dataname]
            od = other.carddata[dataname]
            sd += od
            self.carddata["main"][colname] += other.carddata["main"][colname]
        # --------------------------------------------------------------------
        # deduplicate
        self.dedup()

    def dedup(self):
        """
        parse additional data to find and remove duplicates
        """
        for dataname in self.carddata.keys():
            if dataname == "main":
                continue
            colname = "%sID" % dataname
            data = self.carddata[dataname]
            dedup_data = []  # this will override self.carddata[dataname]
            for ix, dataset in enumerate(data):
                if dataset not in dedup_data:
                    dedup_data.append(dataset)
                    continue
                # dataset is duplicated; change referencing ID
                already_ix = dedup_data.index(dataset)
                self.carddata["main"][colname][ix] = already_ix
            self.carddata[dataname] = dedup_data

    def _offset_repeated_ix(self, colname, offset):
        """offset indices of repeated data in both `main` and
        `self.REPEATED_DATA_NAME`
        """
        old = self.carddata["main"][colname]
        new = [(ix + offset) for ix in old]
        old2new = dict(zip(old, new))
        self.carddata["main"][colname] = [
            old2new[ix] for ix in self.carddata["main"][colname]
        ]

    def append_sparams(self, sparams, **kwargs):
        """fields are provided as `sparams` dictionnary header: value"""
        _sparams = self.DEFAULTS.copy()
        _sparams.update(sparams)
        # reorder fields
        _sparams = {k: _sparams[k] for k in self.fields.values()}
        # if hasattr(self, "clean_sparams"):
        #     self.clean_params(_sparams)
        for header, value in _sparams.items():
            if value is None:
                value = self.DEFAULTS.get(header, value)
            self.carddata["main"][header].append(value)
            # route multiple types fields
            if header in self.MULT_TYPES_FIELDS:
                alternatives = self.MULT_TYPES_FIELDS[header].copy()
                try:
                    # a value (other than None) has been found
                    # we can therefore guess the field name
                    routed_header = alternatives.pop(type(value))
                    self.carddata["main"][routed_header].append(value)
                except KeyError:
                    pass
                # only remains the alternate field
                for typ, routed_header in alternatives.items():
                    self.carddata["main"][routed_header].append(None)
        # =====================================================================
        # repeated fields
        # LOAD: {'starts@': 4, 'fields': ['S', 'L']}
        # PCOMP: {'starts@': 12, 'fields': ['MID', 'T', 'THETA', 'SOUT']}
        # RBE2: {'starts@': 5, 'fields': ['GM'], 'floating': ['ALPHA']}
        # RBE3: {'starts@': 33, 'fields': ['GM', 'CM'], 'floating': ['"ALPHA"', 'ALPHA']}
        # SPC1: {'starts@': 4, 'fields': ['G']}
        # =====================================================================
        if not kwargs:
            return
        repeated_fields_specs = set(self.fields_info()["repeated"])
        # ---------------------------------------------------------------------
        # ensure we do not have floating repetitions
        # for now, only simple single-field repetitions are handled
        if "floating" in repeated_fields_specs:
            raise NotImplementedError("do not handle for now floating repetitions")
        # ---------------------------------------------------------------------
        # repeated MUST finish with 'i'.
        # TODO: discard this test?
        for repeated_input in kwargs:
            if not repeated_input.endswith("i"):
                raise ValueError(f"repeated field {repeated_input} not understood")
        # ---------------------------------------------------------------------
        # get rid of training "i"
        data = {k[:-1]: v for k, v in kwargs.items()}  # 'Gi' -> 'G'
        # and transform a dict of list into a list of dict
        data = transform_dict_of_list(data)
        # ---------------------------------------------------------------------
        # append data
        self.carddata["main"][self.REPEATED_DATA_NAME + "ID"].append(
            self.nb_items() - 1
        )
        self.carddata[self.REPEATED_DATA_NAME].append(data)

    def nb_items(self):
        try:
            return len(self.carddata["main"][next(iter(self.carddata["main"].keys()))])
        except StopIteration:
            return 0

    def append_fields_list(self, fields):
        """fields are provided as text, without the card name"""
        # insert TWO dummy fields such as index in fields list match NASTRAN field
        fields = ["_", "_"] + fields
        # ==================================================================
        # read fixed fields
        # ==================================================================
        if not hasattr(self.fields, "items"):
            raise AttributeError("card %s has no fields items" % self)
        kwargs = {}
        for ix, header in self.fields.items():
            try:
                value = fields[ix]
            except IndexError:
                value = None
            kwargs[header] = value
        self.append_sparams(kwargs)
        return fields

    def append_legacy(self, fields):
        """fields are provided as text, without the card name"""
        # insert TWO dummy fields such as index in fields list match NASTRAN field
        fields = ["_", "_"] + fields
        # ==================================================================
        # read fixed fields
        # ==================================================================
        if not hasattr(self.fields, "items"):
            raise AttributeError("card %s has no fields items" % self)
        for ix, header in self.fields.items():
            try:
                value = fields[ix]
            except IndexError:
                value = None
            if value is None:
                value = self.DEFAULTS.get(header, value)
            self.carddata["main"][header].append(value)
            # route multiple types fields
            if header in self.MULT_TYPES_FIELDS:
                alternatives = self.MULT_TYPES_FIELDS[header].copy()
                routed_header = alternatives.pop(type(value))
                self.carddata["main"][routed_header].append(value)
                # only remains the alternate field
                for typ, routed_header in alternatives.items():
                    self.carddata["main"][routed_header].append(None)
        return fields

    def parse(self, txt, debug=False):
        """simple parser for NASTRAN small-fields bulk entry.

        ⚠ Only to be used for testing purposes! ⚠
        """
        lines = [l.strip() for l in txt.split("\n") if l.strip()]
        if debug:
            import pprint

            pprint.pprint(lines)  # TODO: remove me!
        fields = []
        for l in lines:
            if l.startswith("$"):
                continue
            for fieldID in range(10):
                sub = slice(fieldID * 8, fieldID * 8 + 8)
                fields.append(autoconvert(l[sub].strip()))
        # trim None
        while fields[-1] is None:
            fields.pop(-1)
        if debug:
            print("fields:\n", fields)
        self.append_fields_list(fields[1:])

    def export_data(self):
        """export_data data for serialization"""
        res = deepcopy(self.carddata)
        res["main"] = dict(res["main"])  # defaultdict -> dict
        res["card"] = self.name
        return res

    def resume(self, data):
        """resume card from data,
        such as PCOMP = PCOMP.resume(PCOMP.export_data()) == PCOMP
        """
        self._reset()
        # data = deepcopy(data)
        cardname = data.pop("card")
        if cardname != self.card:
            raise ValueError(
                f'provided cardname "{cardname}" differes from %s' % self.card
            )
        assert cardname == self.card
        for ksrc, ktarget in data.items():
            # keep defaultdict for 'main' data
            if ksrc in self.carddata and hasattr(data[ksrc], "update"):
                self.carddata[ksrc].update(data[ksrc])
            else:
                self.carddata[ksrc] = data[ksrc]

    def __getstate__(self):
        return self.__dict__
        # return {k : v for k, v in self.__dict__.items() if not k.startswith('_cached_')}

    @property
    def array(self):
        """return carddata['main'] as numpy structured arrays"""
        return dic2array(self.carddata["main"])

    def ids(self):
        return set(self.carddata["main"][self.XID_FIELDNAME])

    def query_id(self, value, asview=False, with_loc=False):
        """
        return a tuple (arr, rownbs)
        """
        if isinstance(value, int):
            loc = np.where(self.array[self.XID_FIELDNAME] == value)
            value = self.array[loc]
        else:
            # assume a list has been passed
            mask = np.isin(self.array[self.XID_FIELDNAME], list(value))
            value = self.array[np.where(mask)]
            if with_loc:
                loc = [i for i, v in enumerate(mask) if v]
        if asview:
            value = value.view(asview)
        if with_loc:
            return value, loc
        return value

    def query_id_fast(self, value, columns=None, asview=False):
        """
        fast version of query_id
        return a tuple (arr, rownbs)
        """

        ids = self.carddata["main"][self.XID_FIELDNAME]
        loc = self.query_loc(value)
        if not columns:
            subset = {k: [v[i] for i in loc] for k, v in self.carddata["main"].items()}
        else:
            subset = {k: [self.carddata["main"][k][i] for i in loc] for k in columns}
        if asview:
            subset = np.array(list(subset.values())).T
            if len(loc) == 1:
                return subset[0]
        elif len(loc) == 1:
            subset = {k: v[0] for k, v in subset.items()}
        return subset

    def query_loc(self, value):
        """
        return location in carddata lists
        """
        ids = self.carddata["main"][self.XID_FIELDNAME]
        if isinstance(value, int):
            return [ids.index(value)]
        # assume a list has been passed
        return [ids.index(v) for v in value]

    def _extract_array(self, index, columns, values_type="float64", index_type="int"):
        """transform a numpy structured array into {'data': values, 'index': index, 'columns': columns}"""
        sarr = self.array
        values = rfn.structured_to_unstructured(sarr[columns], dtype=values_type)
        index = rfn.structured_to_unstructured(
            sarr[[index]], dtype=index_type
        ).flatten()
        return {"data": values, "index": index, "columns": columns}

    def subset(self, eids=None):
        """return a card object with selected subset"""
        if self.type != ELEMENT:
            raise AttributeError("gids_header")
        array = self.array
        if eids is None:
            eids = array[self.EID_FIELDNAME].tolist()
        # --------------------------------------------------------------------
        # convert eids to numpy indices
        ixs = np.where(
            np.isin(np.array(self.carddata["main"][self.EID_FIELDNAME]), eids)
        )[0]
        _array = array[ixs]
        newdata = {"main": array2dic(_array, astype=list), "card": self.name}
        for k, vs in self.carddata.items():
            if k == "main":
                continue
            # ----------------------------------------------------------------
            # restrict additional data to relevant eids
            ix = newdata["main"][f"{k}ID"]
            vs = [vs[i] for i in ix]
            newdata[k] = vs
            newdata["main"][f"{k}ID"] = list(range(len(ix)))
        obj = self.__class__(data=newdata)
        return obj

    # ========================================================================
    # cachable attributes previously set by decorators
    # ========================================================================
    @cached_property
    def thk(self):
        """return a dict {eid: thk}"""
        if self.type != ELEMENT or self.dim != "2d" or self.THK_PATTERN is None:
            raise AttributeError("thk")
        # get thicknesses columns
        gids_cols = [
            col for col in self.carddata["main"] if self.THK_PATTERN.match(col)
        ]
        _data = self._extract_array("EID", gids_cols)
        _data["data"] = _data["data"].mean(axis=1)
        _data["name"] = "thk"
        _data.pop("columns")
        return _data

    @cached_property
    def gids_header(self):
        """return a list of Grid IDs headers (eg. ['G1', 'G3', 'G3'])"""
        return self._gids_header()

    @classmethod
    @lru_cache
    def _gids_header(cls):
        """return a list of Grid IDs headers (eg. ['G1', 'G3', 'G3'])"""
        if cls.type != "element":
            raise AttributeError("gids_header")
        gids_cols = [
            col for col in cls.fields_info()["mandatory"] if cls.GIDS_PATTERN.match(col)
        ]
        return gids_cols

    @cached_property
    def mids_header(self):
        """return a list of Grid IDs headers (eg. ['G1', 'G3', 'G3'])"""
        if self.type != PROPERTY:
            raise AttributeError("mids_header")
        mids_cols = [
            col for col in self.carddata["main"] if self.MATS_PATTERN.match(col)
        ]
        return mids_cols

    @cached_property
    def pid2mids(self):
        """return a mapping {PID: frozenset((MIDS))}

        this basic property is OK for basic properties like PROD, PSHELL, *etc.*
        but is not aware of PCOMP, for example. This needs therefore
        to be overriden in those later cases
        """
        if self.type != PROPERTY:
            raise AttributeError("pid2mids")
        ret = {}

        for ix, pid in enumerate(self.carddata["main"]["PID"]):
            ret[pid] = set()
            for midh in self.mids_header:
                mid = self.carddata["main"][midh][ix]
                # print(f'pid {pid}; midh {midh}:: {mid}')
                if mid:  # mid can be `None`; skip this case
                    ret[pid].add(mid)
            ret[pid] = frozenset(ret[pid])
        return ret

    @cached_property
    def _eid2gids_ordered(self):
        if self.type != ELEMENT:
            raise AttributeError("gids_header")
        gids_cols = self.gids_header
        eid2gids = defaultdict(set)
        for ix, eid in enumerate(self.carddata["main"][self.EID_FIELDNAME]):
            eid2gids[eid] = [self.carddata["main"][c][ix] for c in gids_cols]
            if hasattr(self, "eid2gids_complement"):
                # if list == set:
                #     eid2gids[eid] |= self.eid2gids_complement(eid=eid, ix=ix)
                # else:
                nb_meaningfull = len(self.gids_header)
                if nb_meaningfull == 1:
                    msg = f"Only the first ID is meaningfull"
                else:
                    msg = f"Only the first {nb_meaningfull} IDs are meaningfull"
                logging.info(msg)
                eid2gids[eid] += list(self.eid2gids_complement(eid=eid, ix=ix))
        return dict(eid2gids)

    @cached_property
    def _eid2gids(self):
        if self.type != ELEMENT:
            raise AttributeError("gids_header")
        gids_cols = self.gids_header
        eid2gids = defaultdict(set)
        for ix, eid in enumerate(self.carddata["main"][self.EID_FIELDNAME]):
            eid2gids[eid] = set(self.carddata["main"][c][ix] for c in gids_cols)
            if hasattr(self, "eid2gids_complement"):
                eid2gids[eid] |= self.eid2gids_complement(eid=eid, ix=ix)
        return dict(eid2gids)


class RepeatedRowsCard(SimpleCard):
    """Mother Card for cards dfining **ONE SINGLE** 'etc.'"""

    # ------------------------------------------------------------------------
    # tables
    TABLE = None
    REPEATED_ROWS_TABLE = None
    TRAILING_ROWS_TABLE = None
    # ------------------------------------------------------------------------
    # more stuff
    REPEATED_ROWS_NAME = None  # eg. 'stations' for PBEAM
    TRIGGER_REPEATED_ON = str
    SKIP_NEXT_ROW_ON = ()  # e.g. for PBEAM: ('SO', ('YESA', 'NO'))

    def __init__(self, name=None, data=None):
        super().__init__(name=name, data=data)
        self.REPEATED_ROWS_NAME = "{}_{}".format(
            self.name.lower(), self.REPEATED_ROWS_NAME.lower()
        )
        # assert self.repeated is not None
        # set-up an additional container for repeated rows
        self.carddata[self.REPEATED_ROWS_NAME] = []

    def append_fields_list(self, fields):
        fields = super().append_fields_list(fields)  # append regular fields
        # ==================================================================
        # read repeated rows
        # ==================================================================
        # card_subset:  one single PER card
        # example for PBEAM:
        #    {'SO': 'YES', 'X/XB': 0.3, etc...}
        card_subset = dict()
        # card_set: several (at least one) PER card
        # example for PCOMP:
        #   [{'SO': 'YES', 'X/XB': 0.3, etc...},
        #    {'SO': 'YES', 'X/XB': 0.7, etc...}]
        card_set = []
        # --------------------------------------------------------------------
        # analyse REPEATED_ROWS_TABLE and TRAILING_ROWS_TABLE
        _, repfields, _ = parse_table(self.REPEATED_ROWS_TABLE)
        _, trailfields, _ = parse_table(self.TRAILING_ROWS_TABLE)
        # ====================================================================
        # assuming that at least ONE block of repeated rows exists
        # this is True for PBEAM, see [ref.] remark #4
        # ====================================================================
        remaining_fields = fields[get_field(max(self.fields.keys()) + 1) :]
        remaining_fields = ["_", "_"] + remaining_fields
        # nbrepeated_rows = nbrows_by_fields(repfields)
        # offset = nbrepeated_rows * 10  # offset to apply each time we parse a block
        # --------------------------------------------------------------------
        # if some rows need to be skipped...
        _rev_repfields = {
            fieldname: fieldID for fieldID, fieldname in repfields.items()
        }
        try:
            _skipped_fieldID = _rev_repfields[self.SKIP_NEXT_ROW_ON[0]]
            _skipped_on = self.SKIP_NEXT_ROW_ON[1]
        except:
            _skipped_fieldID, _skip_on = None, None
            __import__("pdb").set_trace()
        while isinstance(remaining_fields[2], self.TRIGGER_REPEATED_ON):
            card_subset = dict()
            # number of rows to parse using repfields
            _skipped_field = remaining_fields[_skipped_fieldID]
            if _skipped_field in _skipped_on:  # if SO in ('YESA', 'NO'):
                block_length = 10
            else:
                block_length = 20
            block = remaining_fields[:block_length]
            remaining_fields = remaining_fields[block_length:]
            # build card_subset
            for ix, header in repfields.items():
                try:
                    card_subset[header] = block[ix]
                except IndexError:
                    card_subset[header] = None
            card_set.append(card_subset)
        card_set_ix = len(self.carddata[self.REPEATED_ROWS_NAME])
        self.carddata["main"][self.REPEATED_ROWS_NAME + "ID"].append(card_set_ix)
        self.carddata[self.REPEATED_ROWS_NAME].append(card_set)
        # ====================================================================
        # trailing fields are for the TRAILING_ROWS_TABLE
        # ====================================================================
        if remaining_fields:
            for ix, header in trailfields.items():
                try:
                    self.carddata["main"][header].append(remaining_fields[ix])
                except IndexError:
                    self.carddata["main"][header].append(None)
        return fields

    def to_nastran(
        self, ids=frozenset(), ruler=False, with_default=True, comments=None
    ):
        """
        Convert a collection of cards to 8-chars NASTRAN format.
        RepeatedRowsCard to_nastran()"""

        ffieldsmap, lines, ixid = self._to_nastran(
            ids=ids, ruler=ruler, additional_ruler_lines=["$ + repeated fields..."]
        )

        for ix, id in ixid:
            # initiate a new line and populate a data dict with fixed fields
            try:
                data = self._newline(ffieldsmap, ix, with_default)
            except ValueError as exc:
                logging.critical(f"cannot make new line for {self.name} {ix}")
                raise
            # ----------------------------------------------------------------
            # repeated rows
            _skipped_fieldname, _skipped_on = self.SKIP_NEXT_ROW_ON
            card_set = self.carddata[self.REPEATED_ROWS_NAME][ix]
            _offset = get_field(max([int(k[2:]) for k in data.keys()]) + 1) - 2
            for i, subset in enumerate(card_set):
                if (
                    subset[_skipped_fieldname] in _skipped_on
                ):  # if subset['SO'] in ('YESA', 'NO'):
                    _, repfields, _ = parse_table(self.REPEATED_ROWS_TABLE, linesno=[0])
                else:
                    _, repfields, _ = parse_table(self.REPEATED_ROWS_TABLE)
                offset = _offset + i * _offset
                fieldsmap = DefaultDict()  # , fn10='+', fn11='+')
                fieldsmap.update(
                    {"fn%d" % (k + offset): v for k, v in repfields.items()}
                )
                for fieldcode, fieldname in fieldsmap.items():
                    if fieldcode == "+":
                        continue
                    data[fieldcode] = trans(subset[fieldname])
            # ----------------------------------------------------------------
            # trailing rows

            _, trailfields, _ = parse_table(self.TRAILING_ROWS_TABLE)
            nextfield = get_field(max([int(k[2:]) for k in data.keys()]) + 1) - 2
            trailfields = {k + nextfield: v for k, v in trailfields.items()}
            for fieldcode, fieldname in trailfields.items():
                if fieldcode == "+":
                    continue
                data["fn%d" % fieldcode] = trans(self.carddata["main"][fieldname][ix])
            try:
                lines += fields_to_card(data)
            except:
                __import__("pdb").set_trace()
        return lines


class SimpleCyclingCard(SimpleCard):
    """Mother Card for cards dfining **ONE SINGLE** 'etc.'"""

    REPEATED_DATA_NAME = None  # eg. 'LAYUP'
    TABLE = None
    REPEATED_DEFAULTS = {}

    def __init__(self, name=None, data=None):
        super().__init__(name=name, data=data)
        if data is None:
            self.REPEATED_DATA_NAME = "{}_{}".format(
                self.name.lower(), self.REPEATED_DATA_NAME.lower()
            )
            # assert self.repeated is not None
            # set-up an additional container for cycling data
            self.carddata[self.REPEATED_DATA_NAME] = []

    def to_nastran(
        self, ids=frozenset(), ruler=False, with_default=True, comments=None
    ):
        """
        Convert a collection of cards to 8-chars NASTRAN format.
        SimpleCyclingCard to_nastran()"""

        fieldsmap, lines, ixid = self._to_nastran(ids=ids, ruler=ruler)
        for ix, id in ixid:
            # initiate a new line and populate a data dict with fixed fields
            try:
                data = self._newline(fieldsmap, ix, with_default)
            except ValueError as exc:
                logging.critical(
                    f"cannot make new line for {self.name} {ix} {id=} {fieldsmap=} {with_default=}"
                )
                breakpoint()
                raise
            # repeated data
            # self.carddata['main'][self.REPEATED_DATA_NAME + 'ID']
            rdata_id = self.carddata["main"][self.REPEATED_DATA_NAME + "ID"][ix]
            rdata = self.carddata[self.REPEATED_DATA_NAME][rdata_id]
            # reset field pointer
            field_pnt = self.repeated["starts@"]
            # iterate over repeated data for the current line
            for id, dataset in enumerate(rdata):
                for fieldname in self.repeated["fields"]:
                    field_pnt = get_field(field_pnt)
                    fieldcode = "fn%d" % field_pnt
                    value = dataset[fieldname]
                    if not with_default:
                        value = self._check_default(
                            fieldname, value, default=self.REPEATED_DEFAULTS
                        )
                    data[fieldcode] = trans(value)
                    field_pnt = field_pnt + 1
            # ----------------------------------------------------------------
            # floating (trailing) data
            floating_fields = self.repeated.get("floating", [])
            for fieldname in floating_fields:
                field_pnt = get_field(field_pnt)
                fieldcode = "fn%d" % field_pnt
                value = self.carddata["main"][fieldname][ix]
                if not with_default:
                    value = self._check_default(fieldname, value, default=self.DEFAULTS)
                data[fieldcode] = trans(value)
                field_pnt = field_pnt + 1
            lines += fields_to_card(data)
        return lines

    def append_fields_list(self, fields):
        fields = super().append_fields_list(fields)  # append regular fields
        # ==================================================================
        # read cycling fields, if any...
        # ==================================================================
        # card_subset:  one single PER card
        # example for PCOMP:
        #    {'MID': 2, 'SOUT': 'YES', 'T': 0.023622, 'THETA': 0.0}
        card_subset = dict()
        # card_set: several (at least one) PER card
        # example for PCOMP:
        #   [{'MID': 2, 'SOUT': 'YES', 'T': 0.023622, 'THETA': 0.0},
        #    {'MID': 3, 'SOUT': 'YES', 'T': 0.452756, 'THETA': 0.0},
        #    {'MID': 2, 'SOUT': 'YES', 'T': 0.023622, 'THETA': 0.0}}]
        card_set = []
        # --------------------------------------------------------------------
        # some usefull intermediate variables
        field_nb = self.repeated["starts@"] - 1  # starting field index
        parsed_fields_counter = 0
        # --------------------------------------------------------------
        # we should stop on float if and only if -etc.- is not the last
        # field
        stop_cycling_on_float = "floating" in self.repeated
        while True:  # field_nb < nb_of_fields_to_parse:
            # ----------------------------------------------------------
            # check if end of data to parse is reached
            try:
                # parse next field (if any)
                field_nb += 1
                value = fields[field_nb]
            except:
                # end of job. Well done!
                break
            if stop_cycling_on_float and isinstance(value, float):
                break
            # ----------------------------------------------------------
            # skip continuation fields:
            if field_nb % 10 == 0 or (field_nb - 1) % 10 == 0:
                continue
            # ----------------------------------------------------------
            # get the field name that will be come a fieldname for the `value`
            # already collected
            ix = parsed_fields_counter % len(self.repeated["fields"])
            if ix == 0 and value is None:
                break
            fieldname = self.repeated["fields"][ix]
            parsed_fields_counter += 1
            # ----------------------------------------------------------
            # check default value for repeated stuff
            if value is None:
                value = self.REPEATED_DEFAULTS.get(fieldname, value)
            if fieldname in card_subset:
                # time to change of `card_subset` since we find twice the same key
                card_set.append(card_subset.copy())
                card_subset = {fieldname: value}
            else:
                card_subset[fieldname] = value
        # --------------------------------------------------------------------
        # append last card_subset to set container
        card_set.append(card_subset)
        # check if `card_set` already exists
        container = self.carddata[self.REPEATED_DATA_NAME]
        for i, data in enumerate(container):
            if data == card_set:
                self.carddata["main"][self.REPEATED_DATA_NAME + "ID"].append(i)
                break
        else:
            # nothing existing found
            container.append(card_set)
            self.carddata["main"][self.REPEATED_DATA_NAME + "ID"].append(
                len(container) - 1
            )
        # --------------------------------------------------------------------
        # trailing floating fields (if any)
        floating_fields = self.repeated.get("floating", [])
        for i, fieldname in enumerate(floating_fields):
            try:
                self.carddata["main"][fieldname].append(fields[field_nb + i])
            except:
                # SHALL be in default
                self.carddata["main"][fieldname].append(self.DEFAULTS[fieldname])


class ComplexCard(SimpleCard):
    """Mother Card for specific processing (eg. "PBUSH" and its optional flags)"""

    TABLE = None

    def __init__(self, name=None, data=None):
        super().__init__(name=name, data=data)

    def append_checkin(self, fields):
        """hook triggered right before `append`"""
        return fields

    def append_checkout(self, fields):
        """hook triggered right after `append`"""
        return fields

    def append_fields_list(self, fields):
        fields = self.append_checkin(fields)
        super().append_fields_list(fields)
        self.append_checkout(fields)



import json
import logging
from collections import defaultdict
from copy import deepcopy

import yaml

import nastranio.cards as nio_cards
from nastranio.cards import PROP2ELTS
from nastranio.constants import BOUNDARY

DEFAULT_USERINPUT_TPL = {
    "format": 2,
    "exec": {"SOL": "SESTATIC"},
    "params": {
        "PRGPST": "YES",
        "POST": -1,
        "OGEOM": "NO",
        "AUTOSPC": "YES",
        "K6ROT": 100.0,
        "GRDPNT": 0,
        "SKIPMGG": "YES",
        "CHKGRDS": "NO",
    },
    "cases": {
        "default": {
            "id": -1,
            "TITLE": "LCID1",
            "ECHO": "NONE",
            "DISPLACEMENT(PLOT)": "ALL",
            "$ SPC": 1,
            "$ LOAD": 1,
        }
    },
    "attributes": {
        "materials": {
            1: {"card": "MAT1", "params": {"E": 7e6, "NU": 0.33, "RHO": 5000.5}},
            2: {"card": "MAT1", "params": {"E": 8e6, "NU": 0.33, "RHO": 5000.5}},
        },
        "properties": {
            1: {
                "card": "PBAR",
                "params": {
                    "MID": 1,
                    "A": 50.0,
                    "I1": 18.0,
                    "I2": 18.0,
                    "J": 0.1,
                    "I12": 0.0,
                },
            },
            2: {
                "card": "PBUSH",
                "params": {
                    "K": "K",
                    "K1": 1e4,
                    "K2": 1e4,
                    "K3": 1e4,
                    "K4": 1e4,
                    "K5": 1e4,
                    "K6": 1e4,
                },
            },
            3: {
                "card": "PSHELL",
                "params": {"MID1": 1, "T": 3.0, "MID2": 1, "MID3": 1},
            },
            4: {
                "card": "PCOMP",
                "params": {
                    "NSM": 0.0,
                },
                "layup": [
                    {"MID": 1, "SOUT": "YES", "T": 0.018, "THETA": 0.0},
                    {"MID": 2, "SOUT": "YES", "T": 0.339, "THETA": 0.0},
                    {"MID": 1, "SOUT": "YES", "T": 0.018, "THETA": 5.0},
                ],
            },
        },
        "affectations": {},
        "boundaries": {},
        "loading": {},
    },
}

DEFAULT_BOUNDARY_INPUT = {"card": "SPC1", "params": {"C": "123456"}}
DEFAULT_LOADING_INPUT = {
    "card": "FORCE",
    "params": {
        "CID": 0,
        "F": 1.0,
        "N1": 1.0,
        "N2": 0.0,
        "N3": 0.0,
        "as_sum": False,
        "exclusive": False,
    },
}


def _from_v1_carddata(carddata):
    """
    was [<cardname>, <params>]
    now {"card": <cardname>, "params": <params>}
    """
    try:
        return {"card": carddata[0], "params": carddata[1]}
    except:
        logging.critical(f"cannot import {carddata} from V1")


def from_v1(parameters):
    parameters = deepcopy(parameters)
    parameters["format"] = 2
    for section_name in ("materials", "properties"):
        section = parameters["attributes"][section_name]
        for entry_id, carddata in section.items():
            parameters["attributes"][section_name][entry_id] = _from_v1_carddata(
                carddata
            )
    for section_name in ("boundaries", "loading"):
        section = parameters["attributes"][section_name]
        for sid, group2carddata in section.items():
            for grpname, carddata in group2carddata.items():
                parameters["attributes"][section_name][sid][
                    grpname
                ] = _from_v1_carddata(carddata)
    return parameters


class UserInput:
    def load_parameters(self, parameters):
        convmap = {1: from_v1}
        format = parameters["format"]
        if format != DEFAULT_USERINPUT_TPL["format"]:
            logging.warning(f"importing from deprecated {format=}")
            parameters = convmap[format](parameters)
            self.imported_from = format
        else:
            self.imported_from = None
        self.parameters = parameters
        self.attributes = parameters["attributes"]

    def get_default_template(self):
        return deepcopy(DEFAULT_USERINPUT_TPL)

    def get_default_boundary_input(self):
        return deepcopy(DEFAULT_BOUNDARY_INPUT)

    def get_default_loading_input(self):
        return deepcopy(DEFAULT_LOADING_INPUT)

    def to_json(self, filepath=None, indent=2, sort_keys=False, **kwargs):
        txt = json.dumps(self.parameters, indent=indent, sort_keys=sort_keys, **kwargs)
        if filepath:
            with open(filepath, "w") as fh:
                fh.write(txt)
            return filepath
        return txt

    def to_yaml(self, filepath=None, indent=2, sort_keys=False, **kwargs):
        txt = yaml.dump(self.parameters, sort_keys=sort_keys, **kwargs)
        if filepath:
            with open(filepath, "w") as fh:
                fh.write(txt)
            return filepath
        return txt

    def get_card_data(self, carddata, **kwargs):
        """merge user input with card definition"""
        cardname = carddata["card"]
        params = carddata["params"]
        card = getattr(nio_cards, cardname)
        params = params.copy()
        if "XID" in kwargs:
            kwargs[card.XID_FIELDNAME] = kwargs.pop("XID")
        params.update(kwargs)
        fields_info = card.fields_info()
        mandatory = set(fields_info["mandatory"])
        optional = fields_info["optional"]
        all = mandatory | optional
        unknown = set(params) - all
        missing = mandatory - set(params)
        ok_params = {k: v for k, v in params.items() if k in all}
        ret = {
            "card": card,
            "params": ok_params,
        }
        if hasattr(card, "LOADING_TYPE") and card.LOADING_TYPE is not None:
            ret["over"] = [card.LOADING_TYPE]
        if card.type == BOUNDARY:
            ret["over"] = ["nodes"]
        if missing:
            ret["missing"] = missing
        # ---------------------------------------------------------------------
        # repeated fields
        repeated = fields_info.get("repeated", ())
        # if cardname == "PCOMP":
        #     breakpoint()
        if repeated:
            repeated_data_raw = carddata.get(card.REPEATED_DATA_NAME)
            if repeated_data_raw is None:
                # SPC1 has a REPEATED_DATA_NAME, but not available at this point
                ret["repeated"] = repeated
            else:
                # PCOMP has a REPEATED_DATA_NAME, and data are available
                repeated_data = defaultdict(list)
                for data in repeated_data_raw:
                    for fieldname, value in data.items():
                        repeated_data[fieldname + "i"].append(value)
                ret["repeated"] = dict(repeated_data)
        if unknown:
            ret["unknown"] = {k: v for k, v in params.items() if k in unknown}
        # eg for PCOMP:
        # {
        #  'card': <class 'nastranio.cards.properties.PCOMP'>,
        #  'params': {'NSM': 0.0, 'PID': 4},
        #  'repeated': {'MIDi': [1, 1, 1],
        #            'THETAi': [0.0, 90.0, 0.0],
        #            'Ti': [0.01, 0.05, 0.01]},

        return ret

    def get_materials(self):
        ret = {}
        for mid, carddata in self.attributes["materials"].items():
            ret[mid] = self.get_card_data(carddata, XID=mid)
        return ret

    def get_properties(self):
        """
        return a dictionnary with cards specifications:

        (Pdb++) pp(ret)
        {1: {'card': <class 'nastranio.cards.properties.PBAR'>,
             'params': {'A': 1.0, 'I1': 12.0, 'I2': 13.0, 'MID': 1, 'PID': 1}},
         2: {'card': <class 'nastranio.cards.properties.PSHELL'>,
             'params': {'MID1': 1, 'MID2': 1, 'MID3': 1, 'PID': 2, 'T': 0.08}},
         3: {'card': <class 'nastranio.cards.properties.PSHELL'>,
             'params': {'MID1': 1, 'MID2': 1, 'MID3': 1, 'PID': 3, 'T': 0.03}},
         4: {'card': <class 'nastranio.cards.properties.PCOMP'>,
             'params': {'NSM': 0.0, 'PID': 4},
             'repeated': {'MIDi': [1, 1, 1],
                          'THETAi': [0.0, 90.0, 0.0],
                          'Ti': [0.01, 0.05, 0.01]}}}
        """
        ret = {}
        for pid, carddata in self.attributes["properties"].items():
            ret[pid] = self.get_card_data(carddata, XID=pid)
        return ret

    def get_loading(self):
        ret = defaultdict(list)
        for sid, data in self.attributes["loading"].items():
            for grpname, carddata in data.items():
                _data = self.get_card_data(carddata, XID=sid)
                if grpname != "*":
                    _data["over"].append(grpname)
                ret[sid].append(_data)
        return dict(ret)

    def get_boundaries(self):
        ret = defaultdict(list)
        for sid, data in self.attributes["boundaries"].items():
            for grpname, carddata in data.items():
                _data = self.get_card_data(carddata, XID=sid)
                if grpname != "*":
                    _data["over"].append(grpname)
                ret[sid].append(_data)
        return dict(ret)

    def get_affectations(self):
        ret = {}
        props = self.get_properties()
        for grpname, params in self.attributes["affectations"].items():
            prop_card = props[params["PID"]]["card"]
            cards_choice = PROP2ELTS[prop_card.__name__]
            card_by_type = {}
            for cardname in cards_choice:
                gmsh_elem_type = getattr(nio_cards, cardname).gmsh_eltype
                _data = self.get_card_data(
                    carddata={"card": cardname, "params": params}
                )
                _data["over"] = ["elements", grpname]
                card_by_type[gmsh_elem_type] = _data
            ret[grpname] = card_by_type
        return ret



import json
import logging
import os
import re
from collections import defaultdict, namedtuple
from multiprocessing import Manager, Process
from pathlib import Path

from numtools.serializer import Serializer

from nastranio import cards as cards_mod
from nastranio.constants import (
    BULK,
    CASE,
    COMMENTS,
    EXEC,
    META,
    PARAMS,
    SUMMARY,
    UNKNOWN,
)
from nastranio.decorators import timeit
from nastranio.mesh_api import Mesh
from nastranio.mesh_api.mod import Mod
from nastranio.readers.bulk import read_buffer, read_bulk
from nastranio.utils import calcdiff

# ----------------------------------------------------------------------------
# optional imports...

# fmt: off
try:
    import msgpack
    ISMSGPACK = True
except ImportError:
    ISMSGPACK = False
    logging.debug('`msgpack` not available')

try:
    from deepdiff import DeepDiff
    ISDEEPDIFF = True
except ImportError:
    ISDEEPDIFF = False
    logging.debug('`deepdiff` not available')
# fmt: on


class Registry(Serializer):
    """
    Main public entry point to Nastran Bulk file parsing.
    """

    def __init__(self):
        self.container = None

    def _build_summary(self):
        """(re-)build from scratch a summary dict. Use full when
        registry is manually created
        """
        _summary = defaultdict(set)
        for cardname, card in self.container["bulk"].items():
            _summary[card.type].add(cardname)
            if hasattr(card, "shape"):
                _summary[card.shape].add(cardname)
            if hasattr(card, "dim"):
                _summary[card.dim].add(cardname)
        self.container["summary"] = dict(_summary)

    def diff(self, other, exclude_meta=False):
        return calcdiff((self.container, other.container), exclude_meta=exclude_meta)

    def __eq__(self, other):
        diff = calcdiff((self.container, other.container), exclude_meta=True)
        return len(diff) == 0

    def comments(self):
        """associate cards individuals to comments"""
        # --------------------------------------------------------------------
        # search for FEMAP comments
        if COMMENTS.title not in self.container:
            return {}
        comments = self.container[COMMENTS.title]
        femap_tags = defaultdict(dict)
        regex = re.compile(
            r"^\$\s+Femap\s*(?P<what>(\w+\s+)+)\s*(?P<id>\d+)\s*:\s*(?P<title>.*)$"
        )
        for coms in comments:
            # skip block comment
            for com in coms.split("\n"):
                # if "\n" in com:
                #     continue
                m = regex.match(com)
                if m:
                    femap_tags[m.group("what").replace("with NX Nastran", "").strip()][
                        int(m.group("id"))
                    ] = m.group("title").strip()
        return dict(femap_tags)

    def check_consistancy(self, verbose=True):
        """introspect myself and display some useful infos and warnings"""
        issues = []
        # --------------------------------------------------------------------
        # warning on unknown cards
        ucards = list(self.container.get("summary", {}).get(UNKNOWN, ()))
        if ucards:
            issues.append(("warning", "Unknown cards: %s" % ucards))
        if verbose:
            for level, msg in issues:
                getattr(logging, level)(msg)
        return issues

    def bind_mesh_api(self):
        self.mesh = Mesh()
        self.mesh.set_registry(self, calc_csys=True)

    def bind_mod_api(self, index_elems=("0d", "1d", "2d", "3d")):
        """bind nastranio.mesh_api.mod.Mod class"""
        self.mesh.mod = Mod()
        self.mesh.mod.set_registry(self, index_elems=index_elems)
        logging.info("Modification Module loaded under `reg.mesh.mod`")

    # ========================================================================
    # input / output
    # ========================================================================

    # ------------------------------------------------------------------------
    # NASTRAN BULKFILES

    def from_bulkfile(self, filename, nbprocs="auto", progress=True):
        """Populate a registry from a NASTRAN bulk file"""
        self.container = None
        try:
            fh = filename
            fh.seek(0)
            if nbprocs not in ("auto",) and nbprocs > 1:
                logging.warning("reading filelike object is monoproc")
            containers = read_buffer(fh, progress=progress)
        except AttributeError:
            containers = read_bulk(
                filename=filename, nbprocs=nbprocs, progress=progress
            )
        self.container = containers.pop(0)
        for cont in containers:
            self._merge_container(cont)
        # create self.container['bulk'] dictionnaries:
        self.bind_mesh_api()
        # =====================================================================
        # write all grids to CSYS 0
        # =====================================================================
        modgids = self._translate_grids_to_0()
        if modgids:
            logging.warning("translated {len(modgids)} nodes to CSYS0}")
        self.check_consistancy()

    def _translate_grids_to_0(self):
        """
        ensure IDs described in carddata are in the same order as
        IDs got from  GRID.coords()
        """
        if len(self.container["bulk"]["GRID"]) == 0:
            return set()
        if set(self.container["bulk"]["GRID"].carddata["main"]["CP"]) == set((0,)):
            return set()
        coordsK = self.container["bulk"]["GRID"].coords(asdf=True)  # as defined
        coords0 = self.container["bulk"]["GRID"].coords(
            csysreg=self.mesh.CSys, incsys=-1
        )  # coordinates in CSYS0
        gcdata = self.container["bulk"]["GRID"].carddata["main"]
        ids = gcdata["ID"]
        ids_coords = coords0[0]
        assert list(ids_coords) == ids
        gids_to_modify = coordsK[coordsK["csys"] != 0].index
        for gid in gids_to_modify:
            rowno = self.container["bulk"]["GRID"]._id2rowno[gid]
            gcdata["CP"][rowno] = 0
            xyz0 = coords0[1][rowno]
            gcdata["X1"][rowno] = xyz0[0]
            gcdata["X2"][rowno] = xyz0[1]
            gcdata["X3"][rowno] = xyz0[2]
        self.container["bulk"]["GRID"].clear_caches()
        return set(gids_to_modify)

    def to_nastran(self, filename=None):
        """dumps registry to NASTRAN deck file"""
        bulk = ""
        # ---------------------------------------------------------------------
        # EXEC
        if "exec" in self.container:
            for k, v in self.container["exec"].items():
                bulk += f"{k} {v} \n"
            bulk += "CEND\n"
        # ---------------------------------------------------------------------
        # CASES
        if "cases" in self.container:
            for case_section, data in self.container["cases"].items():
                if case_section != "default":
                    bulk += f"{case_section}\n"
                for k, v in data.items():
                    if k.isupper():
                        bulk += f"  {k} = {v}\n"
        # ---------------------------------------------------------------------
        # BULK (parameters)
        if "param" in self.container or "bulk" in self.container:
            bulk += "BEGIN BULK\n"
        for param, value in self.container.get("params", {}).items():
            bulk += f"PARAM,{param},{value}\n"
        # ---------------------------------------------------------------------
        # BULK (data)
        for cardfamily, cards in self.container.get("bulk", {}).items():
            bulk += "\n".join(cards.to_nastran(comments=self.comments()))
            bulk += "\n"
        # ---------------------------------------------------------------------
        # end
        bulk += "ENDDATA"
        if filename:
            filename = Path(filename).expanduser()
            with open(filename, "w") as fh:
                fh.write(bulk)
            return filename
        return bulk

    # ------------------------------------------------------------------------
    # from / to file

    def from_file(self, fname, fmt="auto"):
        """unserialize registry from a msgpack or json file"""
        if fmt == "auto":
            # guess file type from extension
            fmt = os.path.splitext(fname)[-1][1:]  # "toto.json -> "json"
        self.container = None
        if fmt == "json":
            with open(fname, "r") as fh:
                txt = fh.read()
            self.from_json(txt)

        elif fmt == "pack":
            with open(fname, "rb") as fh:
                txt = fh.read()
            self.from_msgpack(txt)
        else:
            raise ValueError(f"extension/format {fmt} not handled")
        self.check_consistancy()

    def to_file(self, fname=None, fmt="pack"):
        """
        Serialize registry to file. Default is to save to msgpack.

        :param fname: file path
        :type fname: str or `None`
        :param fmt: file's format (and extension)
        :type fmt: str or `None`

        :returns: actual file name
        """
        if not fname:
            # if filename is not provided, save registry besides source file
            # using the passed format `fmt` as extension.
            source = self.meta["source"]
            fname, ext = os.path.splitext(source)
            fname += f".{fmt}"
        if fmt == "json":
            with open(fname, "w") as fh:
                fh.write(self.to_json())
        elif fmt == "pack":
            with open(fname, "wb") as fh:
                fh.write(self.to_msgpack())
        return fname

    def __getattr__(self, attr):
        """
        user-friendly method to access self.container
        """
        # if attr in self.__dict__:
        #     return getattr(self, attr)
        if attr in self.container:
            # logging.debug('proxy to self.container["%s"]', attr)
            return self.container[attr]
        for section, data in self.container.items():
            if attr in data:
                # logging.debug('proxy to self.container["%s"]["%s"]', section, attr)
                return self.container[section][attr]
        raise AttributeError(attr)

    def __getstate__(self):
        return {"container": self.container}

    def __setstate__(self, state):
        self.__dict__ = state
        self.bind_mesh_api()
        # if self.container:
        #     self.check_consistancy()

    def _merge_container(self, container):
        """merge two containers into the same registry.
        use full for multiprocess resuming
        """
        sections = (EXEC, PARAMS, CASE, COMMENTS, BULK, SUMMARY)
        for section in sections:
            if section.title not in container:
                continue
            # if section.title in (EXEC.title, PARAMS.title, CASE.title):
            #     # only bulk and comments should be non-zero
            #     if len(container[section.title]) != 0:
            #         logging.debug("%s::%s", section.title, BULK.title)
            #         raise ValueError("Section {0.title} not empty".format(section))
            if section.title not in self.container:
                self.container[section.title] = {}
            self_d = self.container[section.title]
            other_d = container[section.title]
            logging.debug(f"merge {section.title} (%d items)", len(other_d))
            if section.title in (EXEC.title, PARAMS.title, CASE.title):
                self_d.update(other_d)
            elif section.title == COMMENTS.title:
                # reglar lists
                self_d += other_d
            elif section.title == SUMMARY.title:
                for k in other_d:
                    self_d[k] |= other_d[k]
            elif section.title == BULK.title:
                # append new bulk cards_mod family to self
                for bulk_entry, bulk_entries in container[BULK.title].items():
                    if bulk_entry not in self.container[BULK.title]:
                        self.container[BULK.title][bulk_entry] = bulk_entries
                        logging.debug(
                            'added "%s" (%d items)', bulk_entry, len(bulk_entries)
                        )
                        continue
                    # entry already exists ; we need to append entries
                    else:
                        self.container[BULK.title][bulk_entry].merge(bulk_entries)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""
FEMAP neutral writer
"""

import logging
import os

from jinja2 import Environment, FileSystemLoader
from numtools.intzip import zip_list

TPLPATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")


def _renumber(ints, offset):
    """renumber and return an offseted list of integers and the old->new mapping

    >>> _renumber([1, 2, 3], offset=100)
    ([101, 102, 103], {1: 101, 2: 102, 3: 103})
    >>> _renumber([1, 2, 101], offset=100)
    ([1, 2, 101], {1: 1, 2: 2, 101: 101})
    >>> _renumber([], offset=100)
    ([], {})
    """
    if not ints:
        return [], {}
    _ints = ints.copy()
    if max(ints) < offset:
        ints = [i + offset for i in ints]
    renumbering = dict(zip(_ints, ints))
    return ints, renumbering


_DEFAULT_CLIPPING = (
    #'$COM --- default clipping ---\n'
    "0,0,0,0,0.,0.,\n"  # 4:: coclip:  CC_on, CC_dof, CC_meth, CC_csys, CC_min, CC_max
    "0,0,\n"  # 5:: plcip_meth, plcip_in
    "0,0,\n"  # 6:: + plclip_on, plclip_neg
    "0.,0.,0.,\n"  #     + plclip_base (X, Y, Z)
    "0.,0.,0.,\n"  #     + plclip_norm (X, Y, Z)
    "0,0,\n"  # 6:: + plclip_on, plclip_neg
    "0.,0.,0.,\n"  #     + plclip_base (X, Y, Z)
    "0.,0.,0.,\n"  #     + plclip_norm (X, Y, Z)
    "0,0,\n"  # 6:: + plclip_on, plclip_neg
    "0.,0.,0.,\n"  #     + plclip_base (X, Y, Z)
    "0.,0.,0.,\n"  #     + plclip_norm (X, Y, Z)
    "0,0,\n"  # 6:: + plclip_on, plclip_neg
    "0.,0.,0.,\n"  #     + plclip_base (X, Y, Z)
    "0.,0.,0.,\n"  #     + plclip_norm (X, Y, Z)
    "0,0,\n"  # 6:: + plclip_on, plclip_neg
    "0.,0.,0.,\n"  #     + plclip_base (X, Y, Z)
    "0.,0.,0.,\n"  #     + plclip_norm (X, Y, Z)
    "0,0,\n"  # 6:: + plclip_on, plclip_neg
    "0.,0.,0.,\n"  #     + plclip_base (X, Y, Z)
    "0.,0.,0.,\n"  #     + plclip_norm (X, Y, Z)
    #'$COM --- /end of default clipping ---\n'
)
FEMAP_ENTITIES = {
    # entity: (group_rule_ID, group_entity_ID)
    "gids": (7, 17),
    "eids": (8, 21),
    "mids": (9, 26),
    "pids": (10, 30),
    "pnt_ids": (1, 3),
    "txt_ids": (5, 14),
}


def _dump_group(group, femap_grp_id, com):
    """
    return a string containing the FEMAP neutral definition for the group.
    """
    # ========================================================================
    # group's header
    # ========================================================================
    str_ = "%s,1,0,\n" % int(femap_grp_id)  # ID, refresh?, no_renumbering?
    str_ += "%s\n" % group["title"]  # group title
    str_ += "0,0,0,\n"  # 3:: min layer, max layer, Type of layer usage
    str_ += _DEFAULT_CLIPPING  # clipping stuff
    # ========================================================================
    # rules
    # ========================================================================
    if com:
        str_ += "$COM --- RULES ---\n"
    str_ += "133,\n"  # MAX number of rules
    for entity, ids, femap_attr in zip(
        group["entities"].keys(), group["entities"].values(), FEMAP_ENTITIES.values()
    ):
        if len(ids) == 0:
            continue
        if com:
            str_ += "$COM --- /%d/ %s ---\n" % (femap_attr[1], entity)
        str_ += "%s,\n" % femap_attr[1]
        items_list = zip_list(ids, couple_alone=0)
        nb_lines = len(items_list)
        if nb_lines > 0:
            tpl = "\n".join(nb_lines * ["%d,%d,1,1,"])
            var = tuple((id_ for e in items_list for id_ in e))
            str_ += tpl % var
            str_ += "\n"
        str_ += "-1,-1,-1,-1,\n"
    str_ += "-1,\n"  # End of rule: last record
    # ========================================================================
    # entities
    # ========================================================================
    if com:
        str_ += "$COM --- ENTITIES ---\n"
    str_ += "28,\n"  # MAX number of entities sets
    # Actual group definition
    for entity, ids, femap_attr in zip(
        group["entities"].keys(), group["entities"].values(), FEMAP_ENTITIES.values()
    ):
        if len(ids) == 0:
            continue
        if com:
            str_ += "$COM --- /%d/ %s ---\n" % (femap_attr[0], entity)
        str_ += "%s,\n" % femap_attr[0]
        nb_lines = len(ids)
        tpl = "\n".join(nb_lines * ["%d,"])
        var = tuple(ids)
        str_ += tpl % var
        str_ += "\n-1,\n"
    # ========================================================================
    # that's all, Folks!
    # ========================================================================
    str_ += "-1,\n"
    return str_


class Neutral(object):
    """
    Class handling basic export to FEMAP neutral file
    """

    default_points_data = {
        "type": 0,
        "engine": 0,
        "csys": 0,
        "layer": 1,
        "color": 22,
        "mesh_size": 0.0,
        "propertyID": 0,
        "compositeCurveID": 0,
    }

    def __init__(self, femap_neutral_vers="11.0", database_title=None):
        self.femap_neu_vers = femap_neutral_vers
        self.database_title = database_title
        # self._create_group_block()
        self.layers = {}  # { Layer ID: ??? }
        # { group ID: {'title': <str>,                               # MANDATORY
        #              'entities':{'eids': set(), 'gids': set()},    # MANDATORY
        #              'reference': [grpID1, grpID2, ...]            # OPTIONAL
        #             }}
        self.groups = {}
        self.texts = {}
        self.ids_offset = 1
        logging.info(f"loading template from {TPLPATH}")
        self.tpl_env = Environment(
            autoescape=False,
            loader=FileSystemLoader(TPLPATH),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def _next_id(self, attr):
        ids = set(getattr(self, attr).keys())
        return max(ids, default=self.ids_offset - 1) + 1

    # ========================================================================
    # groups jobs
    # ========================================================================
    def next_group_id(self):
        """return the next available goup id"""
        return self._next_id("groups")

    def add_group(self, title, entities=None, reference=None):
        """
        * 'title': a string defining how the group will be shown in FEMAP
        * 'entities': dictionnary of entities to group ({'eids': ..., 'gids': ...})
        * 'reference': OPTIONAL. iterable of referenced groups
        """
        _entities = dict(zip(FEMAP_ENTITIES.keys(), (set() for i in FEMAP_ENTITIES)))
        if not entities:
            entities = {}
        _entities.update(entities)
        if not reference:
            reference = set()
        nextid = self.next_group_id()
        self.groups[nextid] = {
            "title": title,
            "entities": _entities,
            "reference": reference,
        }
        return nextid

    def refers_groups(self, grpid, grpids):
        """modify an existing group `grpid` to referes some other groups."""
        self.groups[grpid]["reference"] = grpids

    def dump_groups(self, com):
        """write all groups to the neutral file"""
        grpids = sorted(list(self.groups.keys()))
        string = ""
        for grpid in grpids:
            group = self.groups[grpid]
            try:
                string += _dump_group(group, grpid, com)
            except:
                logging.critical(f"cannot dump group ID# {grpid} {group}")
        return string.strip("\n")

    def dump_referenced_groups(self, com):
        msg = ""
        if com:
            msg += "$COM ---[ Referenced Groups ] ---"
        for grpid, group in self.groups.items():
            if "reference" not in group or len(group["reference"]) == 0:
                continue
            msg += "%d\n" % grpid
            msg += ",\n".join(["%d" % g for g in group["reference"]]) + ",\n"
            msg += "-1,\n"
        return msg.strip("\n")

    # ========================================================================
    # layers jobs
    # ========================================================================
    def next_layer_id(self):
        """return the next available goup id"""
        return self._next_id("layers")

    def add_layer(self, title, color=30):
        """add a layer. `layer` is a dictionnary with 2 keys:
        * 'title': a string defining how the group will be shown in FEMAP
        * 'color': OPTIONAL. Defaulted to XXX
        """
        nextid = self.next_layer_id()
        self.layers[nextid] = {"title": title, "color": color}
        return nextid

    def dump_layers(self, com):
        msg = ""
        layers = sorted(self.layers.keys())
        if com:
            msg += "$COM ---[ Layers ] ---"
        for layerID in layers:
            msg += "%d, %d,\n" % (layerID, self.layers[layerID]["color"])
            msg += "%s,\n" % self.layers[layerID]["title"]
        return msg.strip("\n")

    # ========================================================================
    # texts jobs
    # ========================================================================
    def next_text_id(self):
        """return the next available goup id"""
        return self._next_id("texts")

    def add_text(
        self,
        text,
        pointer_loc,
        text_offset=(-1, -1, -1),
        color=124,
        back_color=0,
        bord_color=14,
        font=1,
        layer=1,
        groupID=None,
    ):
        """add a text"""
        data = {
            "text": text,
            "color": color,
            "back_color": back_color,
            "bord_color": bord_color,
            "font": font,
            "layerID": layer,
            "model_positioning": 1,
            "horz_just": 1,  # left
            "vert_just": 0,  # center
            "visible": 1,
            "viewID": 1,
            "draw_pointer": 1,
            "draw_border": 1,
            "pointer_loc": pointer_loc,
            "text_position": [e1 + e2 for e1, e2 in zip(pointer_loc, text_offset)],
            "groupID": groupID,
        }
        nextid = self.next_text_id()
        self.texts[nextid] = data
        if groupID:
            self.groups[groupID]["entities"]["txt_ids"] |= set((nextid,))
        return nextid

    def dump_texts(self):
        """ """
        msg = ""
        txts = self.texts  # .df_text.T.to_dict()
        # prepare template
        tpl = "{txtid:.0f}, {color:.0f}, {back_color:.0f},\
 {bord_color:.0f}, {font:.0f}, {layerID:.0f},\n"
        tpl += "{model_positioning:.0f}, {horz_just:.0f}, {vert_just:.0f},\
 {visible:.0f}, {viewID:.0f}, {draw_pointer:.0f}, {draw_border:.0f},\n"
        tpl += "{tx}, {ty}, {tz},\n"
        tpl += "{px}, {py}, {pz},\n"
        tpl += "{text_lines:.0f},\n"
        tpl += "{text}\n"
        # prepare data
        for txtid, _data in txts.items():
            _data = _data.copy()

            _data["text_lines"] = len(_data["text"].split("\n"))
            _data["txtid"] = txtid
            _loc = _data.pop("text_position")
            _data["tx"] = _loc[0]
            _data["ty"] = _loc[1]
            _data["tz"] = _loc[2]
            _loc = _data.pop("pointer_loc")
            _data["px"] = _loc[0]
            _data["py"] = _loc[1]
            _data["pz"] = _loc[2]
            txt = tpl.format(**_data)
            msg += txt
        return msg.strip("\n")

    # ========================================================================
    # make
    # ========================================================================
    def make(self, com):
        """prepare data for neutral file rendering"""
        data = {
            "com": com,
            # 'database_title': self.database_title,
            "neutral_version": self.femap_neu_vers,
            "layers": self.dump_layers(com=com),
            "groups": self.dump_groups(com=com),
            "referenced_groups": self.dump_referenced_groups(com=com),
            # 'points': self.dump_points(),
            "text": self.dump_texts(),
        }
        return data

    def set_offset(self, offset):
        # apply offset to content
        self.ids_offset = offset
        # --------------------------------------------------------------------
        # renumbering groups
        _, g_renumbering = _renumber(list(self.groups.keys()), offset)
        # self.groups = {g_renumbering[i]: v for i, v in self.groups.items()}
        # also renumber referenced groups
        for grpid, group in self.groups.copy().items():
            newid = g_renumbering[grpid]
            self.groups[newid] = self.groups.pop(grpid)
            assert id(group) == id(self.groups[newid])
            ref = group.get("reference", ())
            if ref:
                logging.info("renumber references for group %d" % newid)
                group["reference"] = [g_renumbering[i] for i in group["reference"]]

        # --------------------------------------------------------------------
        # renumbering layers
        _, l_renumbering = _renumber(list(self.layers.keys()), offset)
        self.layers = {l_renumbering[i]: v for i, v in self.layers.items()}
        # --------------------------------------------------------------------
        # renumbering texts
        _, t_renumbering = _renumber(list(self.texts.keys()), offset)
        for old_text_id, text in self.texts.copy().items():
            new_text_id = t_renumbering[old_text_id]
            self.texts[new_text_id] = self.texts.pop(old_text_id)
            assert id(text) == id(self.texts[new_text_id])
            text["layerID"] = l_renumbering[text["layerID"]]
            if text["groupID"] is not None:
                old_group_id = text["groupID"]
                new_group_id = g_renumbering[text["groupID"]]
                text["groupID"] = new_group_id
                self.groups[new_group_id]["entities"]["txt_ids"].remove(old_text_id)
                self.groups[new_group_id]["entities"]["txt_ids"].add(new_text_id)
        # self.texts = {t_renumbering[i]: v for i, v in self.texts.items()}

    def dump_points(self):
        msg = ""
        pnts = self.model.df_points.T.to_dict()
        pnts = {k + self.ids_offset: v for k, v in pnts.items()}
        string = "{pntid}, {type}, {engine}, {csys}, {layer:.0f}, {color},\
 {mesh_size}, {propertyID}, {compositeCurveID},\n"
        string += "{x}, {y}, {z},\n"
        for pntid, data in pnts.items():
            if pntid == 0:
                continue
            _data = self.default_points_data.copy()
            _data.update(data)
            _data["pntid"] = pntid
            _data["color"] = int(_data["color"]) if _data["color"] else 22
            pnt = string.format(**_data)
            msg += pnt
        return msg.strip("\n")

    # =========================================================================
    # curves
    # =========================================================================
    #    571
    # $com--------------------------
    # 1,28686,0,10004,0,0.,0,0,0,0,0,
    # 0,0,0,
    # 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    # 0.,0.,0.,
    # 0.,0.,0.,0.,0.,0.,
    # 479,480,0,0,0,
    # $com--------------------------
    #    -1

    def get_category_by_grp(self, grp_obj):
        for category, groups in list(self._groups.items()):
            if grp_obj in groups:
                return category

    def write(self, filepath=None, com=False):
        """write the file!"""
        data = self.make(com=com)
        neu = self.tpl_env.get_template("femap_neu.tpl").render(data)
        if not filepath:
            return neu
        with open(filepath, "w") as f:
            f.write(neu)
        return filepath


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    f = Femap_Neu()
    print(80 * "-")
    print(f.write())
    print(80 * "-")
